/*
 *
 *
 */
#include "common.h"
#include "LPC214x.h"
#include "usbhw.h"

/*
 * functions
 */
void vcom_isr_Reset_Event(u_long unused);

/*
 * variables
 */
const INTR_subFUNCs USB_Jump_Table[N_JT];
const INTR_subFUNCs USB_Jump_Table_EP[LPC_N_EP];

/*
 * externs
 */
/* in usbcore.c */
extern void ucore_ResetCore(void);


/*
 * USB Endpoint Events Callback Pointers
 */
const INTR_subFUNCs USB_Jump_Table[N_JT] = {
/*0:USB_JT_POWER  */	NULL  /*vcom_isr_Power_Event*/,
/*1:USB_JT_RESET  */	vcom_isr_Reset_Event,
/*2:USB_JT_SUSPEND*/	NULL  /*vcom_isr_Suspend_Event*/,
/*3:USB_JT_RESUME */	NULL  /*vcom_isr_Resume_Event*/,
/*4:USB_JT_SOF    */	NULL  /*vcom_isr_SOF_Event*/,
/*5:USB_JT_ERROR  */	NULL  /*vcom_isr_Error_Event*/,
};

/*
 *  USB Reset Event Callback
 */
void vcom_isr_Reset_Event(u_long unused)
{
rprintf("uCnc\r\n");
	ucore_ResetCore();
}

/*
 * USB Endpoint Events Callback Pointers
 */
const INTR_subFUNCs USB_Jump_Table_EP[LPC_N_EP] = {
/*00: 0out*/	NULL,		/* control out endpoint, unused */
/*80: 0in*/		NULL,		/* control in endpoint, unused */
/*01: 1out*/	 NULL,		/* interrupt OUT */
/*81: 1in*/		 NULL,		/* interrupt IN */	/*KB*/
/*02: 2out*/	 NULL,		/* bulk OUT */
/*82: 2in*/		 NULL,		/* bulk IN */
/*03: 3out*/	 NULL,		/* iso OUT */
/*83: 3in*/		 NULL,		/* iso IN */
/*04: 4out*/	 NULL,		/* interrupt OUT */
/*84: 4in*/		 NULL,		/* interrupt IN */	/*CTRL*/
/*05: 5out*/	 NULL,		/* bulk OUT */
/*85: 5in*/		 NULL,		/* bulk IN */
/*06: 6out*/	 NULL,		/* iso OUT */
/*86: 6in*/		 NULL,		/* iso IN */
/*07: 7out*/	 NULL,		/* interrupt OUT */
/*87: 7in*/		 NULL,		/* interrupt IN */	/*MS*/
/*08: 8out*/	 NULL,		/* bulk OUT */
/*88: 8in*/		 NULL,		/* bulk IN */
/*09: 9out*/	 NULL,		/* iso OUT */
/*89: 9in*/		 NULL,		/* iso IN */
/*0a:10out*/	 NULL,		/* interrupt OUT */
/*8a:10in*/		 NULL,		/* interrupt IN */
/*0b:11out*/	 NULL,		/* bulk OUT */
/*8b:11in*/		 NULL,		/* bulk IN */
/*0c:12out*/	 NULL,		/* iso OUT */
/*8c:12in*/		 NULL,		/* iso IN */
/*0d:13out*/	 NULL,		/* interrupt OUT */
/*8d:13in*/		 NULL,		/* interrupt IN */
/*0e:14out*/	 NULL,		/* bulk OUT */
/*8e:14in*/		 NULL,		/* bulk IN */
/*0f:15out*/	 NULL,		/* bulk OUT */
/*8f:15in*/		 NULL,		/* bulk IN */
};

#define MOUSE
#define UDEBUG 1

#define SETUP_MASK				0x60	//Used to mask off request type
#define SETUP_CLASS_REQUEST		0x20	//Class Request
//#define GD_DEVICE				0x01	// Get device descriptor: Device
//#define GD_CONFIGURATION		0x02	// Get device descriptor: Configuration
//#define GD_STRING				0x03	// Get device descriptor: String
#define GD_HID	            0x21	// Get descriptor: HID
#define GD_REPORT	         0x22	// Get descriptor: Report

/*
 * defines for HID device
 */
#define UR_GET_REPORT			0x01
#define UR_SET_REPORT			0x09
#define UR_GET_IDLE				0x02
#define UR_SET_IDLE				0x0a
#define     IDLE_TIME_RATIO        4	/* [msec/1] */
#define UR_GET_PROTOCOL			0x03
#define UR_SET_PROTOCOL			0x0b


/*
 * my device's pipe infos (sync with each descriptors in dscr.a51 !)
 */
#ifdef MOUSE
# define IF_HID1     0		/* interface # of HID1 (Keyboard) is 0 */
# define IF_HID2     1		/* interface # of HID2 (Control) is 1 */
# define IF_HID3     2		/* interface # of HID3 (Mouse)  is 2 */
# define N_IFs       3		/* total number of interfaces */
#else /* MOUSE */
# define IF_HID1     0		/* interface # of HID1 (Keyboard) is 0 */
# define IF_HID2     1		/* interface # of HID2 (Control) is 1 */
# define N_IFs       2		/* total number of interfaces */
#endif /* MOUSE */

/* as keyboard */
#define HID1_IN_DATA_LEN   8	/* keycodes --> HOST */
#define HID1_OUT_DATA_LEN  1	/* LED status <-- HOST */

/* as vendor device */
#define HID2_IN_DATA_LEN   2	/* LED status/control_status --> HOST */
#define HID2_OUT_DATA_LEN  9	/* keycodes/mouse/control_status <-- HOST */

/* as mouse */
#define HID3_IN_DATA_LEN   4	/* mouse --> HOST */

/*
 * consts for accessing Nth endpoint data length
 */
const u_char HIDx_in_data_len[N_IFs] = {
	HID1_IN_DATA_LEN,
	HID2_IN_DATA_LEN,
#ifdef MOUSE
	HID3_IN_DATA_LEN,
#endif /* MOUSE */
};

#define HEAD_RESET  0xff	/* you can not pass this as contents !!! */
#define HEAD_LEDS   'L'
#define HEAD_CTRL   'C'		/* misc control for future use */
#define HEAD_KEYS   'K'
#define HEAD_MOUS   'M'

BOOL DR_SetupCommand(void);
BOOL DR_GetDescriptor(void);

extern void ucore_start_DataIn(void);
extern void ucore_start_DataOut(void);
extern void ucore_start_StatusIn(void);
extern void ucore_STALL_EP0(void);

BYTE Protocol[N_IFs];		/* (nothing is influenced by this value) */
BYTE Idle_Rate[N_IFs];
WORD idle_timer[N_IFs];

extern u_long *HIDDscr1, *ReportDscr1, *ReportDscr1Len;
extern u_long *HIDDscr2, *ReportDscr2, *ReportDscr2Len;
#ifdef MOUSE
extern u_long *HIDDscr3, *ReportDscr3, *ReportDscr3Len;
#endif /* MOUSE */

extern /*static*/ u_char *cpipe_data;
extern /*static*/ u_long cpipe_count;
extern /*static*/ volatile int cpipe_out_data_ready;
extern /*static*/ u_long cpipe_out_buff_remain;	/* OUT only */
extern u_char SETUPDAT[8];
extern u_char EP0BUF[64 /*EP0_BUFF_SIZE*/];

/*
 *
 */
BOOL DR_SetupCommand(void)
{
	u_char target_hid;
	BOOL is_hid1;
#ifdef MOUSE
	BOOL is_hid3;
#endif /* MOUSE */
	u_char ans_len, req_len, i;

//#if UDEBUG
//	rprintf(":");
//	for(i=0; i<8; i++)
//		rprintf(" %02x", SETUPDAT[i]);
//	rprintf(" - ");
//#endif /* UDEBUG */

	MON_LED_OFF;	/* turn off LED cause something act */

	/*
	 * feed class request only
	 */
	if((SETUPDAT[0] & SETUP_MASK) != SETUP_CLASS_REQUEST)
		return(TRUE);

	/*
	 * which interface is target ?
	 */
	if((SETUPDAT[5] != 0x00) || (SETUPDAT[4] >= N_IFs))
		goto exit_with_stall;

	target_hid = SETUPDAT[4];
	is_hid1 = (SETUPDAT[4] == IF_HID1);
#ifdef MOUSE
	is_hid3 = (SETUPDAT[4] == IF_HID3);
#endif /* MOUSE */

	/*
	 * I shall not handle so long data
	 */
	if(SETUPDAT[7] != 0x00)
		goto exit_with_stall;

	req_len = SETUPDAT[6];
	ans_len = 0;			/* no need to answer */

	switch(SETUPDAT[1]) {
	case UR_GET_REPORT:			/* 0x01 mandatory */
		rprintf("GET_REPO\n");
			ans_len = HIDx_in_data_len[target_hid];
//			xmemcpy(EP0BUF, INxBUF(target_hid), ans_len);
			break;

	case UR_GET_IDLE:			/* 0x02 optional */
		rprintf("GET_IDLE\n");
			EP0BUF[0] = Idle_Rate[target_hid];
			ans_len = 1;
			break;

	case UR_SET_IDLE:			/* 0x0a optional */
		rprintf("SET_IDLE - %02x\r\n", SETUPDAT[3]);
			/*
			 * ignore if shoter than Polling interval
			 * in endpoint descriptor (10msec)
			 */
			if(SETUPDAT[3] == 0)
				Idle_Rate[target_hid] = 0;	/* indefinite */
			else if(SETUPDAT[3] < 3)
				Idle_Rate[target_hid] = 3;	/* minimum */
			else
				Idle_Rate[target_hid] = SETUPDAT[3];

			idle_timer[target_hid] = 0;
			break;

	case UR_GET_PROTOCOL:		/* 0x03 mandatory if boot dev */
		rprintf("GET_PROT\n");
			EP0BUF[0] = Protocol[target_hid];
			ans_len = 1;
			break;

	case UR_SET_PROTOCOL:		/* 0x0b mandatory if boot dev */
		rprintf("SET_PROT - %02x\r\n", SETUPDAT[2]);
			Protocol[target_hid] = SETUPDAT[2];
			break;

	case UR_SET_REPORT:			/* 0x09 optional */
		rprintf("SET_REPO -");
#ifdef MOUSE
			if(is_hid3)
				goto exit_with_stall;
#endif /* MOUSE */

			if(is_hid1)
				i = HID1_OUT_DATA_LEN;
			else
				i = HID2_OUT_DATA_LEN;

			if(req_len != i)
				goto exit_with_stall;

			/*
			 * Arm endpoint - do it here to clear (after sud avail)
			 * Clear bytecount to allow new data in; also stops NAKing
			 */
//			OUT0BC = 0;
//			while(EP0CS & 0x08)
//				;

			cpipe_data = EP0BUF;
			cpipe_out_buff_remain = 64;
			cpipe_count = i;
			ucore_start_DataOut();
			while(cpipe_out_data_ready == 0)
				;

#if UDEBUG
			rprintf("==");
			for(i=0; i<req_len; i++)
				rprintf(" %02x", EP0BUF[i]);
			rprintf("\r\n");
#endif /* UDEBUG */

//			if(OUT0BC != req_len)
//				goto exit_with_stall;
#if 0
			/*
			 * rewind communication every time
			 * cause I don't know partner's state.
			 * he may NOT be in `RECV_NULL_or_IDLE`
			 * by power-down or suspend of my side.
			 */
				/*
				 * uart1 ̌ɂ鑊ǂȏԂɂ邩
				 * Ȃ̂ŁAMƂɏ߂𑗂邱Ƃɂ
				 */
			set_to_send(HEAD_RESET);
				/*
				 * because here is NOT interrupt layer, I can wait.
				 * so set_to_send() waits when uart send buff is full.
				 * at the result, transaction with host shall be late.
				 */
					/*
					 * ͂ Setup Command ̏́A荞ݏł
					 * Ȃ|[ÖꕔȂ̂ŁAԂĂ
					 * ̂łBȂ̂łłɑMobt@ς
					 *  set_to_send() ֐̓obt@󂭂܂ő҂悤
					 * ܂BN҂̂ƂƁASET_REPORT 
					 * s PC ҂܂B
					 */

			if(is_hid1) {
				set_to_send(HEAD_LEDS);	/* this is for uart1 comm */
				set_to_send(HEAD_LEDS);	/* this is for return data */
			} else {
				/*
				 * data from host have uart1 comminication header
				 * HEAD_KEYS/HEAD_MOUS/HEAD_CTRL on the top
				 */
				if(OUT0BUF[0] == HEAD_KEYS)
					;
				else if(OUT0BUF[0] == HEAD_MOUS) {
					req_len = 1+ HID3_IN_DATA_LEN;	/* reduce uart1 data size */
					filter_mouse_data(&OUT0BUF[1], HID3_IN_DATA_LEN);
				} else if(OUT0BUF[0] == HEAD_CTRL) {
					set_to_send(HEAD_CTRL);	/* this is for uart1 comm */
					req_len = HID2_IN_DATA_LEN;		/* reduce uart1 data size */
				} else
					goto exit_with_stall;
			}

			for(i=0; i<req_len; i++)
				set_to_send(OUT0BUF[i]);
#endif /* 0 */
			break;

	default:
			goto exit_with_stall;
	}

	if(ans_len != 0) {
		if(ans_len > req_len)
			ans_len = req_len;
//		EZUSB_SET_EP_BYTES(IN0BUF_ID, ans_len);
		cpipe_data = EP0BUF;
		cpipe_count = ans_len;
		ucore_start_DataIn();
	} else {
		ucore_start_StatusIn();
	}

	return(FALSE);

	/*
	 * ---------------------------------
	 *    entry for return with error
	 * ---------------------------------
	 */
exit_with_stall:

//	EZUSB_STALL_EP0();		/* Stall End Point 0 */
	ucore_STALL_EP0();
	cpipe_count = 0;

	return(FALSE);
}

/*
 *
 */
BOOL DR_GetDescriptor(void)
{
	BOOL is_hid1;
#ifdef MOUSE
	BOOL is_hid3;
#endif /* MOUSE */
	u_char *dscr;
	u_long len;

	rprintf("GetDesc ");

	if((SETUPDAT[5] != 0x00) || (SETUPDAT[4] >= N_IFs)) {
		rprintf("\r\n");
		return(TRUE);
	}

	is_hid1 = (SETUPDAT[4] == IF_HID1);
#ifdef MOUSE
	is_hid3 = (SETUPDAT[4] == IF_HID3);
#endif /* MOUSE */

	if(SETUPDAT[3]==GD_HID/*0x21*/) {
		rprintf("HID\r\n");

#ifdef MOUSE
		if(is_hid3) {
//			SUDPTRH = MSB((WORD)HIDDscr3);
//			SUDPTRL = LSB((WORD)HIDDscr3);
			dscr = (u_char *)&HIDDscr3;
			len = dscr[/*bLength*/ 0];
		} else
#endif /* MOUSE */
		if(is_hid1) {
//			SUDPTRH = MSB((WORD)HIDDscr1);
//			SUDPTRL = LSB((WORD)HIDDscr1);
			dscr = (u_char *)&HIDDscr1;
			len = dscr[/*bLength*/ 0];
		} else {
//			SUDPTRH = MSB((WORD)HIDDscr2);
//			SUDPTRL = LSB((WORD)HIDDscr2);
			dscr = (u_char *)&HIDDscr2;
			len = dscr[/*bLength*/ 0];
		}

		cpipe_data = dscr;
		cpipe_count = len;
		return(FALSE);
	}

	if(SETUPDAT[3]==GD_REPORT/*0x22*/) {
		rprintf("REPO\r\n");
//	} else if(SETUPDAT[3]==UDESC_PHYSICAL/*0x23*/) {
//		rprintf("PHY\r\n");
	} else {
		rprintf("\r\n");
		return(TRUE);
	}

#ifdef MOUSE
	if(is_hid3) {
//		answer_for_control_in((u_char *)&ReportDscr3, &ReportDscr3Len);
		dscr = (u_char *)&ReportDscr3;
		len = (u_long)&ReportDscr3Len;
	} else
#endif /* MOUSE */
	if(is_hid1) {
//		answer_for_control_in((u_char *)&ReportDscr1, &ReportDscr1Len);
		dscr = (u_char *)&ReportDscr1;
		len = (u_long)&ReportDscr1Len;
	} else {
//		answer_for_control_in((u_char *)&ReportDscr2, &ReportDscr2Len);
		dscr = (u_char *)&ReportDscr2;
		len = (u_long)&ReportDscr2Len;
	}

	cpipe_data = dscr;
	cpipe_count = len;
	return(FALSE);
}
