/*
 *
 *
 */
#include "common.h"
#include "usb.h"
#include "usbhw.h"

/*
 * internal buffer size for endpoint 0
 *   must be greater than max_packet_size described in device desctiptor
 */
#define EP0_BUFF_SIZE    64

/*
 * Max Number of Interfaces
 */
#define MAX_N_IF          4

/*
 * make state machine for CONTROL-PIPE
 */
#define CPS_IDLE     0		/* only SETUP will be accepted */
#define CPS_DATAIN   1		/* expect IN from host */
#define CPS_DATAOUT  2		/* expect OUT from host */
#define CPS_STATOUT  3		/* expect OUT from host */
#define CPS_STATIN   4		/* expect IN from host */
#if 0
[SUDS_IDLE]
 | SETUP(OUT)
 |     WITH DATA
				[SUDS_DATAOUT]
 |         IN->stall
 |         OUT
 |             ->stall if buffer over
 |             ->recv data until specified length--> ep0_data_ready=TRUE
						[SUDS_STATIN]
 |                 IN->send status
 |                 OUT->stall
 |     WITHOUT DATA
				[SUDS_STATIN] <========
 |         IN->send status
 |         OUT->stall
 | SETUP(IN)
 |     WITHOUT DATA
				[ucore_STALL_EP0()]
 |         IN->stall
 |         OUT->stall
 |     WITH DATA
				[SUDS_DATAIN] <=========
 |         OUT->stall
 |         IN->send data (zero-length-packet if needed)
										(datasize<requestsize) &&
										(datasize % packetsize == 0)
					[SUDS_STATOUT]
 |             OUT->send status
 |             IN->stall
#endif /* 0 */

/*
 *
 */
#define EP_CTRL_IN   (0x80 | 0)
#define EP_CTRL_OUT  (0x00 | 0)

/*
 * functions
 */
int ucore_InitialCore(void);
void ucore_ResetCore(void);
void ucore_WakeUp_host(void);
void ucore_feed_SetupCommand(void);
void ucore_SetupCommand(void);
void ucore_start_DataIn(void);
void ucore_start_DataOut(void);
void ucore_start_StatusIn(void);
static int ucore_GetStatus(void);
static int ucore_SetClrFeature(u_long sc);
static int ucore_GetDescriptor(void);
static u_char *ucore_GetStringDscr(u_char* top_string_desc, u_char index);
static u_char *ucore_GetConfigDscr(u_char *top_config_desc, u_char index);
static int ucore_SetConfiguration(void);
static int ucore_SetInterface(void);
void ucore_isr_Ep0(u_long event);
static void ucore_DataInStage(void);
static int ucore_DataOutStage(void);
static void ucore_StatusInStage(void);
static void ucore_StatusOutStage(void);
void ucore_STALL_EP0(void);

/*
 * variables
 */
u_char *USB_DeviceDescriptor;
u_char *USB_ConfigDescriptor;
u_char *USB_StringDescriptor;
u_long max_packet_size_ep0;
           u_short USB_DeviceStatus;
           u_char  USB_Configuration;	/* 0 or configuration number */
/*static*/ u_char  USB_NumInterfaces;	/* of selected configuration */

/*static*/ u_char  ucore_device_addr;
/*static*/ u_long  ucore_bmEP_enabled;
/*static*/ u_long  ucore_bmEP_stalled;

/*static*/ u_char  USB_AltSetting[MAX_N_IF];

/*static*/ int control_pipe_state;
/*static*/ int cpipe_add_zero_packet;
/*static*/ int volatile cpipe_out_data_ready;
/*static*/ u_long cpipe_len_requested;
/*static*/ u_char *cpipe_data;
/*static*/ u_long cpipe_count;
/*static*/ u_long cpipe_out_buff_remain;	/* OUT only */

u_char SETUPDAT[8];
int GotSUD;
/*static*/ u_char EP0BUF[EP0_BUFF_SIZE];

/*
 * externs
 */
/* in dscr.S */
extern u_long *DeviceDscr, *ConfigDscr, *StringDscr;
/* in usbhw.c */
extern int uhw_Init(void);
extern void uhw_WakeUp(void);
extern void uhw_SetAddress(u_char adr);
extern void uhw_Configure(int cfg);
extern void uhw_ConfigEP(u_char *pEPD);
extern void uhw_EnableEP(u_char EPNum);
extern void uhw_DisableEP(u_char EPNum);
extern void uhw_ResetEP(u_char EPNum);
extern void uhw_SetStallEP(u_char EPNum);
extern void uhw_ClrStallEP(u_char EPNum);
extern u_long uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt);
extern void uhw_arm_WriteEP(u_char EPNum);
extern int uhw_ReadEP(u_char EPNum, u_char *pData, u_long DataSize);
extern void uhw_ReadEP_finish(u_char EPNum);
/* in vcomuser.c */
extern void vcom_USB_Configure_Event(void);
extern void vcom_USB_Interface_Event(void);
extern void vcom_USB_Feature_Event(void);
extern int vcom_SetSIOSetup(u_char Cmd, u_char Data);

extern BOOL DR_SetupCommand(void);
extern BOOL DR_GetDescriptor(void);


/*
 * initial valiables about USB core
 */
int ucore_InitialCore(void)
{
	USB_DeviceDescriptor = (u_char *)&DeviceDscr;
	USB_ConfigDescriptor = (u_char *)&ConfigDscr;
	USB_StringDescriptor = (u_char *)&StringDscr;

	max_packet_size_ep0 = USB_DeviceDescriptor[7];

	control_pipe_state = CPS_IDLE;

	/*
	 * initial usbhw.c (max_packet_size_ep0 needed)
	 */
	return(uhw_Init());
}

/*
 * Reset USB Core
 */
void ucore_ResetCore(void)
{
	if(USB_ConfigDescriptor[/*bmAttributes*/ 7] & USB_CONFIG_SELF_POWERED)
		USB_DeviceStatus = USB_GETSTATUS_SELF_POWERED;
	else
		USB_DeviceStatus = 0x00;
/* has USB_GETSTATUS_SELF_POWERED, USB_GETSTATUS_REMOTE_WAKEUP */
	USB_Configuration = 0;
/* has configuration number */
	ucore_device_addr = 0;

#define bmEP_IN(x)  ((1<<16)<<(x))
#define bmEP_OUT(x)  (1<<(x))
#define bmEP_EPnum(n)  (((n)&0x80) ? bmEP_IN((n)&0x0f) : bmEP_OUT((n)&0x0f))
#define DEF_bmEP_ENABLED  (bmEP_IN(0) | bmEP_OUT(0))
	ucore_bmEP_enabled = DEF_bmEP_ENABLED;		/*  */
	ucore_bmEP_stalled = 0x00000000;			/*  */
}

/*
 * USB Remote Wakeup Function
 */
void ucore_WakeUp_host(void)
{
	if(USB_DeviceStatus & USB_GETSTATUS_REMOTE_WAKEUP) {
		uhw_WakeUp();
	}
}

/*
 * take SETUP data out from hardware buffer
 */
void ucore_feed_SetupCommand(void)
{
	uhw_ReadEP(EP_CTRL_OUT, SETUPDAT, sizeof(SETUPDAT));
//	uhw_ReadEP_finish(EP_CTRL_OUT);
// try, how act does he ? ==> status OUT is pend !
	GotSUD = FALSE;

//	SetupPacket.dir   = (SETUPDAT[0] >> 7) & 0x01;
//	SetupPacket.type  = (SETUPDAT[0] >> 5) & 0x03;
//	SetupPacket.recip = (SETUPDAT[0] >> 0) & 0x1f;
//	SetupPacket.bRequest = SETUPDAT[1];
//	SetupPacket.wValue  = SETUPDAT[2] + (SETUPDAT[3] << 8);
//	SetupPacket.wIndex  = SETUPDAT[4] + (SETUPDAT[5] << 8);
//	SetupPacket.wLength = SETUPDAT[6] + (SETUPDAT[7] << 8);

	cpipe_len_requested = SETUPDAT[6] + (SETUPDAT[7] << 8);
}

/*
 * device request parser
 */
#if 1
void ucore_SetupCommand(void)
{
  if(DR_SetupCommand())
	switch((SETUPDAT[0] >> 5) & 0x03 /*type*/) {
	case REQUEST_STANDARD:
			switch(SETUPDAT[1]) {
			case USB_REQUEST_GET_STATUS:
rprintf("GET_STAT\r\n");
					if(!ucore_GetStatus()) {
						goto stall_i;
					}
					ucore_start_DataIn();
					break;

			case USB_REQUEST_CLEAR_FEATURE:
rprintf("CLR_FEAT\r\n");
					if(!ucore_SetClrFeature(0)) {
						goto stall_i;
					}
					ucore_start_StatusIn();
//					vcom_USB_Feature_Event();
					break;

			case USB_REQUEST_SET_FEATURE:
rprintf("SET_FEAT\r\n");
					if(!ucore_SetClrFeature(1)) {
						goto stall_i;
					}
					ucore_start_StatusIn();
//					vcom_USB_Feature_Event();
					break;

			case USB_REQUEST_SET_ADDRESS:
rprintf("SET_ADDR\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_DEVICE:
							ucore_device_addr = 0x80 | SETUPDAT[2];
							ucore_start_StatusIn();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_GET_DESCRIPTOR:
rprintf("GET_DESC\r\n");
				if(DR_GetDescriptor())
					if(!ucore_GetDescriptor()) {
						goto stall_i;
					}
					ucore_start_DataIn();
					break;

			case USB_REQUEST_SET_DESCRIPTOR:
rprintf("SET_DESC\r\n");
					ucore_STALL_EP0();
					cpipe_count = 0;
					break;

			case USB_REQUEST_GET_CONFIGURATION:
rprintf("GET_CONF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_DEVICE:
							EP0BUF[0] = USB_Configuration;
							cpipe_data = EP0BUF;
							cpipe_count = 1;
							ucore_start_DataIn();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_SET_CONFIGURATION:
rprintf("SET_CONF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_DEVICE:
							if(!ucore_SetConfiguration()) {
								goto stall_i;
							}
							ucore_start_StatusIn();
//							vcom_USB_Configure_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_GET_INTERFACE:
rprintf("GET_INTF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_INTERFACE:
							if(
								(USB_Configuration != 0) &&
								(SETUPDAT[4] < USB_NumInterfaces)
							) {
								EP0BUF[0] = USB_AltSetting[SETUPDAT[4]];
								cpipe_data = EP0BUF;
								cpipe_count = 1;
								ucore_start_DataIn();
							} else {
								goto stall_i;
							}
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_SET_INTERFACE:
rprintf("SET_INTF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_INTERFACE:
							if(!ucore_SetInterface()) {
								goto stall_i;
							}
							ucore_start_StatusIn();
//							vcom_USB_Interface_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			default:
					goto stall_i;
			}
			break;

	case REQUEST_CLASS:
#if UDR_NEED_CLASS_SUPPORT
			switch(SETUPDAT[0] & 0x1f /*recip*/) {
			case REQUEST_TO_INTERFACE:
					//break;
			default:
					goto stall_i;
			}
#else /* UDR_NEED_CLASS_SUPPORT */
			goto stall_i;
#endif /* UDR_NEED_CLASS_SUPPORT */

	case REQUEST_VENDOR:
#if UDR_NEED_VENDOR_SUPPORT
			UARTSettingCmd = SetupPacket.bRequest & 0xFF;
			UARTSettingData = SetupPacket.wValue & 0xFF;
			if(!vcom_SetSIOSetup( UARTSettingCmd, UARTSettingData )) {
				goto stall_i;
			}
			ucore_start_DataIn();
#endif /* UDR_NEED_VENDOR_SUPPORT */
			break;

	default:
stall_i:
			ucore_STALL_EP0();
			cpipe_count = 0;
			break;
	}
rprintf("@\r\n");
}
#else /* n */

BOOL DR_SetConfiguration(void);
BOOL DR_GetConfiguration(void);
BOOL DR_SetInterface(void) ;
BOOL DR_GetInterface(void) ;
BOOL DR_GetStatus(void);
BOOL DR_ClearFeature(void);
BOOL DR_SetFeature(void);

BOOL Selfpwr = FALSE;
BOOL Rwuen = FALSE;

BYTE Configuration = 0;		// Current configuration
BYTE AlternateSetting = 0;	// Alternate settings

/*
 *
 */
void ucore_SetupCommand(void)
{
#define CONFIG_DSCR		0x02		// Descriptor type: Configuration

#define SC_GET_STATUS			0x00	// Setup command: Get Status
#define SC_CLEAR_FEATURE		0x01	// Setup command: Clear Feature
#define SC_RESERVED				0x02	// Setup command: Reserved
#define SC_SET_FEATURE			0x03	// Setup command: Set Feature
#define SC_SET_ADDRESS			0x05	// Setup command: Set Address
#define SC_GET_DESCRIPTOR		0x06	// Setup command: Get Descriptor
#define SC_SET_DESCRIPTOR		0x07	// Setup command: Set Descriptor
#define SC_GET_CONFIGURATION	0x08	// Setup command: Get Configuration
#define SC_SET_CONFIGURATION	0x09	// Setup command: Set Configuration
#define SC_GET_INTERFACE		0x0a	// Setup command: Get Interface
#define SC_SET_INTERFACE		0x0b	// Setup command: Set Interface
#define SC_SYNC_FRAME			0x0c	// Setup command: Sync Frame
#define SC_ANCHOR_LOAD			0xa0	// Setup command: Anchor load
	
#define GD_DEVICE				0x01	// Get device descriptor: Device
#define GD_CONFIGURATION		0x02	// Get device descriptor: Configuration
#define GD_STRING				0x03	// Get device descriptor: String
#define GD_HID	            0x21	// Get descriptor: HID
#define GD_REPORT	         0x22	// Get descriptor: Report

#define GS_DEVICE				0x80	// Get Status: Device
#define GS_INTERFACE			0x81	// Get Status: Interface
#define GS_ENDPOINT				0x82	// Get Status: End Point

#define FT_DEVICE				0x00	// Feature: Device
#define FT_ENDPOINT				0x02	// Feature: End Point

	u_char *dscr_ptr;

#if 0
  if(DR_SetupCommand())
#endif /* 0 */
	switch(SETUPDAT[1])
	{
		case SC_SET_ADDRESS:
			switch(SETUPDAT[0])
			{
				case 0x00:				// Device
					ucore_device_addr = 0x80 | SETUPDAT[2];
					ucore_start_StatusIn();
					break;
				default:
					ucore_STALL_EP0();		// Stall End Point 0
					break;
			}
			break;

		case SC_GET_DESCRIPTOR:						// *** Get Descriptor
//			if(DR_GetDescriptor())
				switch(SETUPDAT[3])			
				{
					case GD_DEVICE:				// Device
//						SUDPTRH = MSB(pDeviceDscr);
//						SUDPTRL = LSB(pDeviceDscr);
							cpipe_data = (u_char *)USB_DeviceDescriptor;
							cpipe_count = USB_DeviceDescriptor[0];
							ucore_start_DataIn();
						break;
					case GD_CONFIGURATION:			// Configuration
//						if(dscr_ptr = (void *)EZUSB_GetConfigDscr(SETUPDAT[2]))
if(dscr_ptr = ucore_GetConfigDscr(USB_ConfigDescriptor, SETUPDAT[2]))
						{
//							SUDPTRH = MSB(dscr_ptr);
//							SUDPTRL = LSB(dscr_ptr);
							cpipe_data = dscr_ptr;
							cpipe_count = dscr_ptr[2] + (dscr_ptr[3]<<8);
							ucore_start_DataIn();
						}
						else
							ucore_STALL_EP0(); 	// Stall End Point 0
						break;
					case GD_STRING:				// String
//						if(dscr_ptr = (void *)EZUSB_GetStringDscr(SETUPDAT[2]))
if(dscr_ptr = ucore_GetStringDscr(USB_StringDescriptor, SETUPDAT[2]))
						{
//							/*
//							 * I extract codes below as a function
//							 */
//							answer_for_control_in(dscr_ptr,
//											((STRINGDSCR *)dscr_ptr)->length);
							cpipe_data = dscr_ptr;
							cpipe_count = dscr_ptr[0];
							ucore_start_DataIn();
						}
						else 
							ucore_STALL_EP0();	// Stall End Point 0
						break;
					default:				// Invalid request
						ucore_STALL_EP0();		// Stall End Point 0
				}
			break;
		case SC_GET_INTERFACE:						// *** Get Interface
			DR_GetInterface();
			break;
		case SC_SET_INTERFACE:						// *** Set Interface
			DR_SetInterface();
			break;
		case SC_SET_CONFIGURATION:					// *** Set Configuration
			DR_SetConfiguration();
			break;
		case SC_GET_CONFIGURATION:					// *** Get Configuration
			DR_GetConfiguration();
			break;
		case SC_GET_STATUS:						// *** Get Status
			if(DR_GetStatus())
				switch(SETUPDAT[0])
				{
					case GS_DEVICE:				// Device
						EP0BUF[0] = ((BYTE)Rwuen << 1) | (BYTE)Selfpwr;
						EP0BUF[1] = 0;
//						EZUSB_SET_EP_BYTES(IN0BUF_ID,2);
						cpipe_data = EP0BUF;
						cpipe_count = 2;
						ucore_start_DataIn();
						break;
					case GS_INTERFACE:			// Interface
						EP0BUF[0] = 0;
						EP0BUF[1] = 0;
//						EZUSB_SET_EP_BYTES(IN0BUF_ID,2);
						cpipe_data = EP0BUF;
						cpipe_count = 2;
						ucore_start_DataIn();
						break;
					case GS_ENDPOINT:			// End Point
//						EP0BUF[0] = EPIO[EPID(SETUPDAT[4])].cntrl & bmEPSTALL;
						EP0BUF[0] = (((0)));
						EP0BUF[1] = 0;
//						EZUSB_SET_EP_BYTES(IN0BUF_ID,2);
						cpipe_data = EP0BUF;
						cpipe_count = 2;
						ucore_start_DataIn();
						break;
					default:				// Invalid Command
						ucore_STALL_EP0();		// Stall End Point 0
				}
			break;
		case SC_CLEAR_FEATURE:						// *** Clear Feature
//			if(DR_ClearFeature())
				switch(SETUPDAT[0])
				{
					case FT_DEVICE:				// Device
						if(SETUPDAT[2] == 1)
						{
							Rwuen = FALSE; 		// Disable Remote Wakeup
							ucore_start_StatusIn();
						}
						else
							ucore_STALL_EP0();	// Stall End Point 0
						break;
					case FT_ENDPOINT:			// End Point
						if(SETUPDAT[2] == 0)
                  {
//							EZUSB_UNSTALL_EP( EPID(SETUPDAT[4]) );
//                     EZUSB_RESET_DATA_TOGGLE( SETUPDAT[4] );
						uhw_ClrStallEP(SETUPDAT[4] & 0x8f);
							ucore_start_StatusIn();
                  }
						else
							ucore_STALL_EP0();	// Stall End Point 0
						break;
					default:
						ucore_STALL_EP0();	// Stall End Point 0
				}
			break;
		case SC_SET_FEATURE:						// *** Set Feature
//			if(DR_SetFeature())
				switch(SETUPDAT[0])
				{
					case FT_DEVICE:				// Device
						if(SETUPDAT[2] == 1)
						{
							Rwuen = TRUE;		// Enable Remote Wakeup
							ucore_start_StatusIn();
						}
						else
							ucore_STALL_EP0();	// Stall End Point 0
						break;
					case FT_ENDPOINT:			// End Point
						if(SETUPDAT[2] == 0)
						{
//							EZUSB_STALL_EP( EPID(SETUPDAT[4]) );
							uhw_SetStallEP(SETUPDAT[4] & 0x8f);
							ucore_start_StatusIn();
						}
						else
							ucore_STALL_EP0();	 // Stall End Point 0
						break;
					default:
						ucore_STALL_EP0();	 // Stall End Point 0
				}
			break;
		default:							// *** Invalid Command
#if 0
			/*
			 * Ƀx_[R}h`Ȃ̂ŁAvOZk̂
			 * B
			 */
			if(DR_VendorCmnd())
#endif /* 0 */
				ucore_STALL_EP0();				// Stall End Point 0
	}
}

BOOL DR_SetConfiguration(void)	// Called when a Set Configuration command is received
{
rprintf("SetConf\r\n");

	Configuration = SETUPDAT[2];
	ucore_start_StatusIn();
	return(TRUE);				// Handled by user code
}

BOOL DR_GetConfiguration(void)	// Called when a Get Configuration command is received
{
rprintf("GetConf\r\n");

	EP0BUF[0] = Configuration;
//	EZUSB_SET_EP_BYTES(IN0BUF_ID,1);
	cpipe_data = EP0BUF;
	cpipe_count = 1;
	ucore_start_DataIn();
	return(TRUE);				// Handled by user code
}

BOOL DR_SetInterface(void) 		// Called when a Set Interface command is received
{
rprintf("SetIf\r\n");

	AlternateSetting = SETUPDAT[2];
	ucore_start_StatusIn();
	return(TRUE);				// Handled by user code
}

BOOL DR_GetInterface(void) 		// Called when a Set Interface command is received
{
rprintf("GetIf\r\n");

	EP0BUF[0] = AlternateSetting;
//	EZUSB_SET_EP_BYTES(IN0BUF_ID,1);
	cpipe_data = EP0BUF;
	cpipe_count = 1;
	ucore_start_DataIn();
	return(TRUE);				// Handled by user code
}

BOOL DR_GetStatus(void)
{
rprintf("GetStat\r\n");

	return(TRUE);
}

BOOL DR_ClearFeature(void)
{
rprintf("ClrFeat\r\n");

	return(TRUE);
}

BOOL DR_SetFeature(void)
{
rprintf("SetFeat\r\n");

	return(TRUE);
}

#endif /* n */

/*
 * initiate data IN
 */
void ucore_start_DataIn(void)
{
	/*
	 * round data length by request length
	 */
	if(cpipe_count > cpipe_len_requested) {
		cpipe_count = cpipe_len_requested;
	}

	/*
	 * I expect IN from the host
	 */
	control_pipe_state = CPS_DATAIN;
	cpipe_add_zero_packet = FALSE;		/* for IN data */

	/*
	 * parse the request finished, but keep OUT buffer filled,
	 * no OUT packet should come until status-out phase
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * initiate first data-in packet
	 */
	ucore_DataInStage();
}

/*
 * initiate data OUT
 */
void ucore_start_DataOut(void)
{
	/*
	 * round data length by request length
	 */
	if(cpipe_count > cpipe_len_requested) {
		cpipe_count = cpipe_len_requested;
	}

	/*
	 * I expect OUT from the host
	 */
	control_pipe_state = CPS_DATAOUT;
	cpipe_out_data_ready = FALSE;		/* for OUT data */

	/*
	 * parse the request finished, make out buffer free
	 */
	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * wait for OUT packet will come
	 */
}

/*
 * initiate status IN
 */
void ucore_start_StatusIn(void)
{
	/*
	 * I expect IN from the host
	 */
	control_pipe_state = CPS_STATIN;

	/*
	 * parse the request finished, but keep OUT buffer filled,
	 * no OUT packet should come until status-out phase
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * initiate status-in packet
	 */
	ucore_StatusInStage();
}

/*
 * Get Status USB Request
 */
static int ucore_GetStatus(void)
{
	u_long n, m;

	switch(SETUPDAT[0] & 0x1f /*recip*/) {
	case REQUEST_TO_DEVICE:
			EP0BUF[0] = (USB_DeviceStatus>>0) & 0xff;
			EP0BUF[1] = (USB_DeviceStatus>>8) & 0xff;
			break;

	case REQUEST_TO_INTERFACE:
			if((USB_Configuration != 0) && (SETUPDAT[4] < USB_NumInterfaces)) {
				EP0BUF[0] = 0;
				EP0BUF[1] = 0;
			} else {
				return(FALSE);
			}
			break;

	case REQUEST_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			m = bmEP_EPnum(n);
			if(
				((USB_Configuration != 0) || ((n & 0x0F) == 0)) &&
				(ucore_bmEP_enabled & m)
			) {
				EP0BUF[0] = (ucore_bmEP_stalled & m) ? 1 : 0;
/*/* shall sync with stall operation ! */
				EP0BUF[1] = 0;
			} else {
				return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}

	cpipe_data = EP0BUF;
	cpipe_count = 2;
	return(TRUE);
}

/*
 * Set/Clear Feature USB Request
 *   pass sc as: 0 for Clear, 1 for Set
 */
static int ucore_SetClrFeature(u_long sc)
{
	u_long n, m;

	switch(SETUPDAT[0] & 0x1f /*recip*/) {
	case REQUEST_TO_DEVICE:
			if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == USB_FEATURE_REMOTE_WAKEUP) {
				if(sc) {
					USB_DeviceStatus |=  USB_GETSTATUS_REMOTE_WAKEUP;
				} else {
					USB_DeviceStatus &= ~USB_GETSTATUS_REMOTE_WAKEUP;
				}
			} else {
				return(FALSE);
			}
			break;

	case REQUEST_TO_INTERFACE:
			return(FALSE);

	case REQUEST_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			m = bmEP_EPnum(n);
			if(
				(USB_Configuration != 0) &&
				((n & 0x0F) != 0) &&
				(ucore_bmEP_enabled & m)
			) {
				if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == USB_FEATURE_ENDPOINT_STALL) {
					if(sc) {
						uhw_SetStallEP(n);
						ucore_bmEP_stalled |=  m;
					} else {
						uhw_ClrStallEP(n);
						ucore_bmEP_stalled &= ~m;
					}
				} else {
					return(FALSE);
				}
			} else {
				return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}
	return(TRUE);
}

/*
 * Get Descriptor USB Request
 */
static int ucore_GetDescriptor(void)
{
	u_char *dscr;
	u_long len;

	switch(SETUPDAT[0] & 0x1f /*recip*/) {
	case REQUEST_TO_DEVICE:
			switch(SETUPDAT[3]) {
			case USB_DEVICE_DESCRIPTOR_TYPE:
rprintf(".device\r\n");
					dscr = (u_char *)USB_DeviceDescriptor;
					len = dscr[/*bLength*/ 0];
					break;

			case USB_CONFIGURATION_DESCRIPTOR_TYPE:
rprintf(".config\r\n");
					dscr = (u_char *)USB_ConfigDescriptor;
					dscr = ucore_GetConfigDscr(USB_ConfigDescriptor, SETUPDAT[2]);
					if(dscr == NULL) {
						return(FALSE);
					}
					len = dscr[2] + (dscr[3]<<8);	/* wTotalLength */
					break;

			case USB_STRING_DESCRIPTOR_TYPE:
rprintf(".string\r\n");
					dscr = ucore_GetStringDscr(USB_StringDescriptor, SETUPDAT[2]);
					if(dscr == NULL) {
						return(FALSE);
					}
					len = dscr[/*bLength*/ 0];		/* bLength */
					break;

			default:
					return(FALSE);
			}
			break;

	case REQUEST_TO_INTERFACE:
			switch(SETUPDAT[3]) {
			default:
					return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}

	cpipe_data = dscr;
	cpipe_count = len;
//	if(cpipe_count > len) {
//		cpipe_count = len;
//	}
//
	return(TRUE);
}

/*
 * search Nth config descriptor
 */
static u_char *ucore_GetConfigDscr(u_char *top_config_desc, u_char index)
{
	u_char *dscr;

	dscr = top_config_desc;

	while(dscr[1 /*type*/] == USB_CONFIGURATION_DESCRIPTOR_TYPE) {
		if(index == 0)
			return(dscr);
		index--;
		dscr += dscr[2] + (dscr[3]<<8);		/* wTotalLength */
	}

	return(NULL);
}

/*
 * search Nth string descriptor
 */
static u_char *ucore_GetStringDscr(u_char* top_string_desc, u_char index)
{
	u_char *dscr;

	dscr = top_string_desc;

	while(dscr[1 /*type*/] == USB_STRING_DESCRIPTOR_TYPE) {
		if(index == 0)
			return(dscr);
		index--;
		dscr += dscr[0 /*length*/];
	}

	return(NULL);
}

/*
 * Set Configuration USB Request
 */
static int ucore_SetConfiguration(void)
{
	u_char *dscr;
	u_long alt, n, m;

	if(SETUPDAT[2]) {
		/*
		 * loop for all configuration descriptors
		 *   don't forget to add '0x00' at
		 *   the end of configuration descriptors !
		 */
		dscr = (u_char *)USB_ConfigDescriptor;

		while(dscr[/*bLength*/ 0]) {
			switch(dscr[/*bDescriptorType*/ 1]) {
			case USB_CONFIGURATION_DESCRIPTOR_TYPE:
					/*
					 * search requested configuration descriptor set
					 */
					if(dscr[/*bConfigurationValue*/ 5] == SETUPDAT[2]) {
						/*
						 * match
						 */
						USB_Configuration = SETUPDAT[2];
						USB_NumInterfaces = dscr[/*bNumInterfaces*/ 4];

						/*
						 * initial misc params/hardwares
						 */
						for(n=0; n<MAX_N_IF; n++) {
							USB_AltSetting[n] = 0;
						}
						for(n=1; n<16; n++) {
							if(ucore_bmEP_enabled & bmEP_OUT(n)) {
								uhw_DisableEP(n);
							}
							if(ucore_bmEP_enabled & bmEP_IN(n)) {
								uhw_DisableEP(n | 0x80);
							}
						}
						ucore_bmEP_enabled = DEF_bmEP_ENABLED;
						ucore_bmEP_stalled = 0x00000000;
						uhw_Configure(TRUE);

						if(dscr[/*bmAttributes*/ 7] & USB_CONFIG_SELF_POWERED) {
							USB_DeviceStatus |=  USB_GETSTATUS_SELF_POWERED;
						} else {
							USB_DeviceStatus &= ~USB_GETSTATUS_SELF_POWERED;
						}
					} else {
						/*
						 * point next configuration descriptor
						 */
						dscr += dscr[2] + (dscr[3]<<8);	/* wTotalLength */
						continue;
					}
					break;

			case USB_INTERFACE_DESCRIPTOR_TYPE:
					alt = dscr[/*bAlternateSetting*/ 3];
					break;

			case USB_ENDPOINT_DESCRIPTOR_TYPE:
					if(alt == 0) {
						/*
						 * enable endpoints about default setting
						 */
						n = dscr[/*bEndpointAddress*/ 2] & 0x8F;
						m = bmEP_EPnum(n);
						ucore_bmEP_enabled |= m;
						uhw_ConfigEP(dscr);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
					}
					break;
			}

			dscr += dscr[/*bLength*/ 0];
		}
	} else {
		/*
		 * configuration canceled
		 */
		USB_Configuration = 0;

		for(n=1; n<16; n++) {
			if(ucore_bmEP_enabled & bmEP_OUT(n)) {
				uhw_DisableEP(n);
			}
			if(ucore_bmEP_enabled & bmEP_IN(n)) {
				uhw_DisableEP(n | 0x80);
			}
		}
		ucore_bmEP_enabled  = DEF_bmEP_ENABLED;
		ucore_bmEP_stalled  = 0x00000000;
		uhw_Configure(FALSE);
	}

	if(USB_Configuration == SETUPDAT[2]) {
		return(TRUE);
	} else {
		return(FALSE);
	}
}

/*
 * Set Interface USB Request
 */
static int ucore_SetInterface(void)
{
	int set;
	u_char *dscr;
	u_char ifn, alt, old;
	u_long msk, n, m;

	/*
	 * am I configured ?
	 */
	if(USB_Configuration == 0)
		return(FALSE);

	set = FALSE;

	/*
	 * loop for all configuration descriptors
	 *   don't forget to add '0x00' at
	 *   the end of configuration descriptors !
	 */
	dscr = (u_char *)USB_ConfigDescriptor;

	while(dscr[/*bLength*/ 0]) {
		switch(dscr[/*bDescriptorType*/ 1]) {
		case USB_CONFIGURATION_DESCRIPTOR_TYPE:
				/*
				 * search requested configuration descriptor set
				 */
				if(dscr[/*bConfigurationValue*/ 5] != USB_Configuration) {
					/*
					 * point next configuration descriptor
					 */
					dscr += dscr[2] + (dscr[3]<<8);		/* wTotalLength */
					continue;
				}
				break;

		case USB_INTERFACE_DESCRIPTOR_TYPE:
				/*
				 * is this descriptor match for requested ?
				 */
				ifn = dscr[/*bInterfaceNumber*/ 2];
				alt = dscr[/*bAlternateSetting*/ 3];
				msk = 0;
				if((ifn == SETUPDAT[4]) && (alt == SETUPDAT[2])) {
					set = TRUE;
					old = USB_AltSetting[ifn];
					USB_AltSetting[ifn] = alt;
				}
				break;

		case USB_ENDPOINT_DESCRIPTOR_TYPE:
				/*
				 * it this descriptor is match for requested ?
				 */
				if(ifn == SETUPDAT[4]) {
					n = dscr[/*bEndpointAddress*/ 2] & 0x8F;
					m = bmEP_EPnum(n);
					if(alt == SETUPDAT[2]) {
						/*
						 * enable endpoints about this setting
						 */
						ucore_bmEP_enabled |=  m;
						ucore_bmEP_stalled &= ~m;
						uhw_ConfigEP(dscr);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
						msk |= m;
					} else if((alt == old) && ((msk & m) == 0)) {
						/*
						 * disble endpoints about old setting...
						 */
						ucore_bmEP_enabled &= ~m;
						ucore_bmEP_stalled &= ~m;
						uhw_DisableEP(n);
							/*
							 * we scans descriptors by 1 pass.
							 * is this loop logic correct ???
							 */
					}
				}
				break;
//		default:
//				set = FALSE;
//				break;
		}

		/*
		 * point next descriptor in the configuration descriptor set
		 */
		dscr += dscr[/*bLength*/ 0];
	}

	return(set);
}

/* ==================================================================== */

/*
 * USB endpoint 0 interrupt handler
 */
void ucore_isr_Ep0(u_long event)
{
	switch(event) {
	case USB_EVT_SETUP:
rprintf("EVT_SETUP\r\n");
			GotSUD = TRUE;
			break;

	case USB_EVT_OUT:
rprintf("EVT_OUT\r\n");
			if(control_pipe_state == CPS_DATAOUT) {
				if(cpipe_count != 0) {
					if(ucore_DataOutStage() == 0) {
						if(cpipe_count == 0) {
							/*
							 * make status-in ready
							 */
//							control_pipe_state = CPS_STATIN;
//							ucore_StatusInStage();
							control_pipe_state = CPS_IDLE;
							cpipe_out_data_ready = TRUE;
						} else {
							/*
							 * permit to receive next OUT packet
							 */
							uhw_ReadEP_finish(EP_CTRL_OUT);
						}
					} else {
						/* ignore this packet */
						uhw_ReadEP_finish(EP_CTRL_OUT);
						control_pipe_state = CPS_IDLE;
						ucore_STALL_EP0();
					}
				} else {
					/* ignore this packet */
					uhw_ReadEP_finish(EP_CTRL_OUT);
					control_pipe_state = CPS_IDLE;
					ucore_STALL_EP0();
				}
			}
			else if(control_pipe_state == CPS_STATOUT) {
				/*
				 * always ignore 'status-OUT-has-come'
				 *   (following next SETUP shall be always
				 *    'Packet over-written', don't mind)
				 */
				//ucore_StatusOutStage();
				//control_pipe_state = CPS_IDLE;
			}
			else {
				control_pipe_state = CPS_IDLE;
				ucore_STALL_EP0();
			}
			break;

	case USB_EVT_IN:
rprintf("EVT_IN\r\n");
			if(control_pipe_state == CPS_DATAIN) {
				if((cpipe_count != 0) || cpipe_add_zero_packet) {
					ucore_DataInStage();
				} else {
					control_pipe_state = CPS_STATOUT;
					/*
					 * permit to receive OUT packet
					 */
					uhw_ReadEP_finish(EP_CTRL_OUT);
				}
			}
			else if(control_pipe_state == CPS_STATIN) {
				/*
				 * 'status-in-finished' here, nothing to do.
				 * only about SET_ADDR request, update device address here.
				 */
				if(ucore_device_addr & 0x80) {
					ucore_device_addr &= 0x7F;
					uhw_SetAddress(ucore_device_addr);
				}
				control_pipe_state = CPS_IDLE;
			}
			else {
				control_pipe_state = CPS_IDLE;
				ucore_STALL_EP0();
			}
			break;
	}
}

/*
 * USB endpoint 0 interrupt handler for data IN 
 */
static void ucore_DataInStage(void)
{
	u_long cnt;

	/*
	 * calc data length for this packet
	 */
	if(cpipe_count > max_packet_size_ep0) {
		cnt = max_packet_size_ep0;
	} else {
		cnt = cpipe_count;
	}
rprintf("DataIn(): [n=%d] %02x %02x..\r\n", cnt, cpipe_data[0], cpipe_data[1]);

	/*
	 * write the data into the buffer
	 */
	uhw_WriteEP(EP_CTRL_IN, cpipe_data, cnt);

	/*
	 *
	 */
	if(cpipe_add_zero_packet) {
		/*
		 * already do this operation, reset this flag
		 */
		cpipe_add_zero_packet = FALSE;
	} else {
		/*
		 * advance pointers for next operation
		 */
		cpipe_data += cnt;
		cpipe_count -= cnt;
		cpipe_len_requested -= cnt;

		/*
		 * add zero-length-packet if the data length is just fit
		 * to packet size while the host expect more.
		 */
		if(
			(cnt == max_packet_size_ep0) &&
			(cpipe_count == 0) &&
			(cpipe_len_requested != 0)
		) {
			cpipe_add_zero_packet = TRUE;
		}
	}

	/*
	 * at last, arm the endpoint
	 *   (to prevent 'transfer-finished' interrupt among above process)
	 */
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * USB endpoint 0 interrupt handler for data OUT
 */
static int ucore_DataOutStage(void)
{
	int cnt;

	/*
	 * take received data out
	 */
	cnt = uhw_ReadEP(EP_CTRL_OUT, cpipe_data, cpipe_out_buff_remain);
rprintf("DataOut(): [n=%d] %02x %02x..\r\n", cnt, cpipe_data[0], cpipe_data[1]);

	if(cnt < 0) {
		return(-1);		/* buffer overflow */
	}

	/*
	 * advance pointers for next operation
	 */
	cpipe_data += cnt;
	cpipe_count -= cnt;
//	cpipe_len_requested -= cnt;
	cpipe_out_buff_remain -= cnt;

	/*
	 * at last, free the endpoint
	 *   (to prevent next 'data-received' interrupt among above process)
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	return(0);
}

/*
 * USB endpoint 0 interrupt handler for status IN
 */
static void ucore_StatusInStage(void)
{
rprintf("StatusInStage()\r\n");
	uhw_WriteEP(EP_CTRL_IN, NULL, 0);
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * USB endpoint 0 interrupt handler for status OUT
 */
static void ucore_StatusOutStage(void)
{
rprintf("StatusOutStage()\r\n");
	uhw_ReadEP(EP_CTRL_OUT, EP0BUF, 0);
	uhw_ReadEP_finish(EP_CTRL_OUT);
}

/*
 * stall both End Point 0 in/out
 */
void ucore_STALL_EP0(void)
{
	uhw_SetStallEP(EP_CTRL_IN);		/* for IN */
	uhw_SetStallEP(EP_CTRL_OUT);	/* for OUT */
rprintf("STALL_EP0()\r\n");
		/*
		 * the stall bits are automatically cleared when next SETUP received
		 */
}
